//
// Implementation of Echofon database client
//
// Copyright (c) 2009 Kazuho Okui / naan studio, Inc. ALL RIGHTS RESERVED.
//

var Cc = Components.classes;
var Ci = Components.interfaces;
var CLASS_ID = Components.ID("0eef4ea0-ce51-11dd-ad8b-0800200c9a66");
var CLASS_NAME = "Echofon Database";
var CONTRACT_ID = "@echofon.com/echofon-database;1";
var ECHOFON_UUID = "twitternotifier@naan.net";
var SQLITE_DB_FILENAME = "echofon_2.0.sqlite";

function EchofonDatabase()
{
  this.wrappedJSObject = true;
  this._conn = null;
}

EchofonDatabase.prototype = {

  QueryInterface: function(aIID) {
    // add any other interfaces you support here
    if (!aIID.equals(Ci.nsISupports) &&
        !aIID.equals(Ci.nsIObserver) &&
        !aIID.equals(Ci.nsIEchofonDatabase))
        throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  observe: function(subject, topic, data) {
  },

  openDatabase: function(user_id) {
    this.recipient_id = user_id;
    this.file = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
    this.file.append('echofon_' + user_id + '.sqlite');

    if (!this.file.exists()) {
      Components.utils.import("resource://echofon/EchofonUtils.jsm");
      EchofonUtils.dumpStackTrace();
      this.log("Database file does not exist for user " + user_id);
      this.createDatabase();
    }

    try {
      var storageService = Cc["@mozilla.org/storage/service;1"].getService(Ci.mozIStorageService);
      this._conn = storageService.openDatabase(this.file);
    }
    catch (e) {
      this.log("Failed to open database.(" + e.message + ") try re-create database file...");
      this.createDatabase();
      try {
        var storageService = Cc["@mozilla.org/storage/service;1"].getService(Ci.mozIStorageService);
        this._conn = storageService.openDatabase(this.file);
      }
      catch (e) {
        this.log("Failed to open database: " + e.message);
        return 0;
      }
    }

    // For debugging
    //
//    this.exec("DELETE FROM statuses WHERE id > (SELECT id FROM statuses ORDER BY id DESC LIMIT 1 OFFSET 5)");
//    this.exec("DELETE FROM home WHERE id > (SELECT id FROM home ORDER BY id DESC LIMIT 1 OFFSET 5)");
//    this.exec("DELETE FROM mentions WHERE id > (SELECT id FROM home ORDER BY id DESC LIMIT 1 OFFSET 5)");
//    this.exec("DELETE FROM mentions");
//    this.exec("DELETE FROM direct_messages WHERE id > (SELECT id FROM direct_messages ORDER BY id DESC LIMIT 1 OFFSET 10)");
//    this.exec("DELETE FROM direct_messages");
//    this.exec("DELETE FROM threads");

    return 1;
  },

  createDatabase: function() {
//    var prompt = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
//    prompt.alert(null, "Echofon", "Create database!");
//    Components.utils.import("resource://echofon/EchofonUtils.jsm");
//    EchofonUtils.dumpStackTrace();
    try {
      this.file.remove(false);
    }
    catch (e) {}

    var defaultFile;
    try {
      // For Firefox 3.6
      var manager = Cc["@mozilla.org/extensions/manager;1"].getService(Ci.nsIExtensionManager);
      defaultFile = manager.getInstallLocation(ECHOFON_UUID).getItemLocation(ECHOFON_UUID);
      defaultFile.append("defaults");
    }
    catch (e) {
      // For Firefox 4
      var prop = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties);
      defaultFile = prop.get("ProfD", Ci.nsIFile);
      defaultFile.append("extensions");
      defaultFile.append(ECHOFON_UUID);
      defaultFile.append("defaults");

      // For XULRunner
      if (!defaultFile.exists()) {
        defaultFile = prop.get("DefRt", Ci.nsIFile);
      }
    }
    defaultFile.append(SQLITE_DB_FILENAME);
    defaultFile.copyTo(this.file.parent, this.file.leafName);
    this.log("Copy database file for user " + this.recipient_id);
  },

  exec: function(sql) {
    try {
      this._conn.executeSimpleSQL(sql);
    }
    catch (e) {
      this.log("Failed to execute query: (" + sql + ") " + this._conn.lastErrorString);
    }
  },

  prepare: function(sql) {
    var stmt;
    try {
      stmt = this._conn.createStatement(sql);
    }
    catch (e) {
      this.log("Failed to prepare statement: (" + sql + ") " + this._conn.lastErrorString);
    }
    return stmt;

  },

  close: function() {
    this._conn.close();
  },

  begin: function() {
    try {
      this._conn.executeSimpleSQL("BEGIN");
    }
    catch (e) {
      this.log("Failed to begin transaction: " + this._conn.lastErrorString);
    }
  },

  commit: function() {
    try {
      this._conn.executeSimpleSQL("COMMIT");
    }
    catch (e) {
      this.log("Failed to commit transaction: " + this._conn.lastErrorString);
    }
  },

  rollback: function() {
    try {
      this._conn.executeSimpleSQL("ROLLBACK");
    }
    catch (e) {
      this.log("Failed to rollback transaction: " + this._conn.lastErrorString);
    }
  },

  insert: function(obj) {
    var sql = "INSERT INTO " + obj._tablename + " (";
    for (var col_name in obj) {
      sql += col_name + ",";
    }
    sql += ") VALUES (";
    for (var value in obj) {
      sql += "?,";
    }
    sql += ")";
    dump(sql);
  },

  logError: function(msg) {

    if (this._conn.lastError >= 100) return;

    msg = msg + " (" + this._conn.lastErrorString + " (" + this._conn.lastError + ") / db " + this.recipient_id + ") ";
    this.log(msg);

    var pref = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefBranch);
    pref.setBoolPref("extensions.twitternotifier.clearDB", true);

    var param = {"state": "internalError", "data": msg + " Please restart Firefox."};

    Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService).notifyObservers(null, "echofon-status", param.toSource());
  },

  log: function(msg) {
    if (this._console == null) {
        this._console = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
    }
    this._console.logStringMessage(msg);
    dump(msg + "\n");
  },

  lastError: function() {
    return this._conn.lastError;
  },

  lastErrorString: function() {
    return this._conn.lastErrorString;
  }
};


//=================================================
// Note: You probably don't want to edit anything
// below this unless you know what you're doing.
//

// Factory
var EchofonDatabaseFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new EchofonDatabase()).QueryInterface(aIID);
  }
};

// Module
var EchofonDatabaseModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);
  },

  getClassObject: function(aCompMgr, aCID, aIID)  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return EchofonDatabaseFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

//module initialization
function NSGetModule(aCompMgr, aFileSpec) { return EchofonDatabaseModule; }

function NSGetFactory(cid) {
  if (!CLASS_ID.equals(cid)) throw Components.results.NS_ERROR_FACTORY_NOT_REGISTERED;
  return EchofonDatabaseFactory;
}
