//
// Implementation of Echofon account manager
//
// Copyright (c) 2010 Kazuho Okui / naan studio, Inc. ALL RIGHTS RESERVED.
//

var EXPORTED_SYMBOLS = ["AccountManager"];

var JSON = Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON);
var pref = Components.classes['@mozilla.org/preferences-service;1'].getService(Components.interfaces.nsIPrefService).getBranch("extensions.twitternotifier.");

var gManager = null;

function AccountManager()
{
  this.accounts = JSON.decode(pref.getCharPref("accounts"));
}

AccountManager.prototype = {
  allAccounts: function() {
    return this.accounts;
  },

  numAccounts: function() {
    var n = 0;
    for (var i in this.accounts) {
      ++n;
    }
    return n;
  },

  getPrimaryAccount: function() {
    for (var i in this.accounts) {
      return i;
    }
    return 0;
  },

  get: function(user_id) {
    var acct = null
    if (user_id) {
      acct = this.accounts[user_id];
    }
    else {
      acct = this.accounts[pref.getIntPref("activeUserId")];
    }
    if (acct) {
      this.addMethods(acct);
    }
    return acct;
  },

  add: function(acct) {
    this.accounts[acct.user_id] = acct;
    this.addMethods(acct);
    this.save();
    return acct;
  },

  remove: function(user_id) {
    delete this.accounts[user_id];
    this.save();
  },

  save: function() {
    var val = JSON.encode(this.accounts);
    pref.setCharPref("accounts", val);
  },

  addMethods: function(acct) {
    acct.setValue = function(name, value) {
      this[name] = value;
    }

    acct.save = function() {
      gManager.save();
    }
    acct.destroy = function() {
      gManager.removeAccount(this.user_id);
    }
  }
};

AccountManager.instance = function() {
  if (gManager == null) {
    gManager = new AccountManager();
  }
  return gManager;
}
