//
// Copyright (c) 2010 Kazuho Okui / naan studio, Inc. ALL RIGHTS RESERVED.
//

var EXPORTED_SYMBOLS = ["EchofonSync"];

const Cc = Components.classes;
const Ci = Components.interfaces;

const IDLE_TIME_INTERVAL = 15 * 60; // 15min
//const SYNC_SERVER = "http://localhost:8080";
const SYNC_SERVER = "https://echofonsync.appspot.com";
const DEFAULT_SYNC_INTERVAL = 10 * 60 * 1000;

var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);

Components.utils.import("resource://echofon/EchofonUtils.jsm");
Components.utils.import("resource://echofon/Models.jsm");
Components.utils.import("resource://echofon/Account.jsm");

function EchofonSync()
{
  try {
    this._signer = Cc['@naan.net/twitterfox-sign;1'].getService(Ci.nsITwitterFoxSign);
  }
  catch (e) {}

  var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  obs.addObserver(this, "quit-application-granted", false);

  var idleService = Cc["@mozilla.org/widget/idleservice;1"].getService(Ci.nsIIdleService);
  idleService.addIdleObserver(this, IDLE_TIME_INTERVAL);

  this._sync = JSON.decode(EchofonUtils.pref().getCharPref("sync"));

  this.resetTimer();

  this._isIdle = false;
  this._lastSync = Date.now();
}

EchofonSync.prototype = {

  isSynced: function(user_id) {
    if (!this._sync[user_id] || !this._sync[user_id].key) {
      return null;
    }
    return this._sync[user_id].key;
  },

  getSyncData: function(user_id) {
    if (!this._sync[user_id]) {
      return null;
    }
    return this._sync[user_id];
  },

  disableSync: function(user_id) {
    delete this._sync[user_id].key;
    delete this._sync[user_id].sync;
    this.save();
  },

  resetTimer: function() {
    if (this._syncTimer) {
      this._syncTimer.cancel();
    }
    this._syncTimer = EchofonUtils.setDelayTask(this.syncInterval(), this, "updateSyncPoint", null, Ci.nsITimer.TYPE_REPEATING_SLACK);
  },

  syncInterval: function() {
    try {
      for (var i in this._sync) {
        var interval = this._sync[i].config.interval;
        if (interval < 3) interval = 3;
        return interval * 60 * 1000;
      }
    }
    catch (e) {
    }
    return DEFAULT_SYNC_INTERVAL;
  },

  registerUser: function(user_id, screen_name, callback) {
    var req = Cc["@echofon.com/echofon-http-request;1"].createInstance(Ci.nsIEchofonHttpRequest);

    var target = this;
    req.onload    = function(p) {target.onLoadRegister(req,  user_id, callback)}
    req.onerror   = function(p) {target.onErrorRegister(req, user_id, callback);}
    req.ontimeout = function(p) {target.onErrorRegister(req, user_id, callback);}

    req.setURL(SYNC_SERVER + "/desktop/register");
    var param = "twitter_id=" + user_id + "&screen_name=" + screen_name;
    req.setPostData(param);
    req.setRequestHeader('X-Echofon-Client', 'Firefox');
    var sig = this.addSignature(param);
    req.setRequestHeader('X-Echofon-Sig', sig);
    req.setTimeout(15 * 1000);
    req.asyncOpen();

    return req;
  },

  onLoadRegister: function(req, user_id, callback) {
    if (Number(req.status) == 200) {
      var key = req.getResponseHeader("X-Echofon-Sess");
      if (this._sync[user_id]) {
        this._sync[user_id].key = key;
      }
      else {
        this._sync[user_id] = {key:key};
      }

      var resp = JSON.decode(req.responseText);
      this._sync[user_id].config = resp.config;
      this._sync[user_id].sync   = resp.sync_str;
      this._sync[user_id].mute   = resp.mute;
      this._lastSync = Date.now();

      this.save();
      callback['userDidRegister'](true);
    }
    else {
      callback['userDidRegister'](false);
    }
  },

  onErrorRegister: function(req, user_id, callback) {
    callback['userDidRegister'](false);
  },

  convertType: function(type) {
    if (type == 'home') return 'timeline';
    if (type == 'mentions') return 'replies';
    if (type == 'messages') return 'messages';
    return null;
  },

  markRead: function(id, user, type) {
    type = this.convertType(type);
    if (!type) return;

    if (!this._sync[user] || !this._sync[user]['sync']) {
      return;
    }

    try {
      if (!this._sync[user]['sync'][type] || DBManager._64bitsub(id, this._sync[user]['sync'][type]) > 0) {
        this._sync[user]['sync'][type] = id;
      }
      this.save();
    }
    catch (e) {}
  },

  isUnread: function(user, type, id) {
    if (!this._sync[user]) return true;
    type = this.convertType(type);
    if (!type) return true;

    try {
      if (DBManager._64bitsub(id, this._sync[user]['sync'][type]) <= 0) {
        return false;
      }
    }
    catch (e) {}

    return true;
  },

  isMuted: function(user, tweet) {
    if (tweet.type == 'lists' || tweet.type == 'search') return false;
    if (!this._sync[user] || !this._sync[user].mute) return false;
    var mute = this._sync[user].mute;
    // check user
    for (var i in mute.user) {
      if (tweet.user && tweet.user.screen_name.toLowerCase() == mute.user[i].toLowerCase()) {
        return true;
      }
    }
    // check application
    if (mute.app && mute.app.length && tweet.source && tweet.source.match(/<a href\=\"([^\"]*)\"[^>]*>(.*)<\/a>/)) {
      var app = RegExp.$2;
      for (var i in mute.app) {
        if (app == mute.app[i]) {
          return true;
        }
      }
    }

    if (!mute.hashtag || mute.hashtag.length == 0 || !tweet.entities || tweet.entities.hashtags.length == 0) return false;

    // check hashtag
    for (var i in mute.hashtag) {
      for (var j in tweet.entities.hashtags) {
        if (('#' + tweet.entities.hashtags[j].text.toLowerCase()) == mute.hashtag[i].toLowerCase()) {
          return true;
        }
      }
    }

    return false;
  },

  isMutedUser: function(user_id, screen_name) {
    if (!this._sync[user_id]) return false;
    if (!this._sync[user_id].mute) return false;
    var m = this._sync[user_id].mute.user;
    if (!m) return false;

    for (var i in m) {
      if (m[i] == screen_name) return true;
    }
    return false;
  },

  mute: function(user_id, value, type) {
    if (!this._sync[user_id]) {
      this._sync[user_id] = {};
    }
    if (!this._sync[user_id].mute) {
      this._sync[user_id].mute = {app:[], hashtag:[], user:[]};
    }

    var m = this._sync[user_id].mute[type];
    if (!m) {
      this._sync[user_id].mute[type] = [];
      m = this._sync[user_id].mute[type];
    }
    for (var i in m) {
      if (m[i] == value) return false;
    }
    m.push(value);

    this._sync[user_id].mute.updated_at = parseInt(Date.now() / 1000);
    this.save();

    if (!this._sync[user_id].key) return true;

    // send to sync server.
    var array = ['key=' + this._sync[user_id].key];
    array.push('mute=' + JSON.encode(this._sync[user_id].mute));

    var params = array.join('&');
    var req = this.postSyncData(user_id, '/sync/mute', params);
    return true;
  },

  unmute: function(user_id, value, type) {
    if (!this._sync[user_id]) return;
    if (!this._sync[user_id].mute) return;

    var m = this._sync[user_id].mute[type];

    var index = -1;

    for (var i in m) {
      if (m[i] == value) {
        index = i;
        break;
      }
    }
    if (index == -1) return;

    m.splice(i, 1);
    this._sync[user_id].mute.updated_at = parseInt(Date.now() / 1000);
    this.save();

    if (this._sync[user_id].key) {
      var array = ['key=' + this._sync[user_id].key];
      array.push('mute=' + JSON.encode(this._sync[user_id].mute));

      var params = array.join('&');
      var req = this.postSyncData(user_id, '/sync/mute', params);
    }
  },

  getSyncPoint: function(user, callback) {
    if (!this._sync[user] || !this._sync[user].key) {
      return false;
    }

    var param = 'key=' + this._sync[user]['key'];
    if (EchofonUtils.pref().getBoolPref("preventPush")) {
      param += '&is_active=1';
    }

    var req = Cc["@echofon.com/echofon-http-request;1"].createInstance(Ci.nsIEchofonHttpRequest);

    var target = this;
    req.onload    = function() {target.onLoadSync(req, user, callback)}
    req.onerror   = function() {target.onErrorSync(req, user, callback);}
    req.ontimeout = function() {target.onErrorSync(req, user, callback);}

    req.setURL(SYNC_SERVER + '/sync/?' + param);
    var sig = this.addSignature(param);
    req.setRequestHeader('X-Echofon-Sig', sig);
    req.setTimeout(5 * 1000);
    req.asyncOpen();

    return true;
  },

  postSyncData: function(user, path, params) {
    var req = Cc["@echofon.com/echofon-http-request;1"].createInstance(Ci.nsIEchofonHttpRequest);

    var target = this;
    req.onload    = function() {target.onLoadSync(req, user)}
    req.onerror   = function() {target.onErrorSync(req, user);}
    req.ontimeout = function() {target.onErrorSync(req, user);}

    req.setURL(SYNC_SERVER + path);
    req.setPostData(params);
    var sig = this.addSignature(params);
    req.setRequestHeader('X-Echofon-Sig', sig);
    req.setTimeout(15 * 1000);
    req.asyncOpen();

    return req;
  },

  updateSyncPoint: function(no_async) {
    var acct = AccountManager.instance().get();
    if (!acct) return false;
    var user = acct.user_id;

    if (!this._sync[user] || !this._sync[user].key) {
      return false;
    }

    if (this._isIdle > 0) {
      this._isIdle++;
      // reduce sync interval to 30 minutes if computer is inactive
      if (this._isIdle % 3 != 0) return false;
    }

    var array = [];
    for (var k in this._sync[user]['sync']) {
      array.push(k + '=' + this._sync[user]['sync'][k]);
    }
    array.push('key=' + this._sync[user].key);
    array.push('client=TwitterFox ' + EchofonUtils.get_version());

    if (EchofonUtils.pref().getBoolPref("preventPush")) {
      var is_active = (this._isIdle == 0) ? true : false;
      array.push('is_active=' + ((is_active) ? '1' : '0'));
    }

    var params = array.join('&');
    var req = this.postSyncData(user, '/sync/', params, no_async);

    return true;
  },

  addSignature: function(data) {
    var sig = '';
    try {
      sig = this._signer.sign(data);
    }
    catch (e) {
      logWarn("Can't load sync module. This platform is not supported.");
      return "";
    }
    var toHexString = function(charCode) { return ("0" + charCode.toString(16)).slice(-2); };
    return [toHexString(sig.charCodeAt(i)) for (i in sig)].join("");
  },

  onLoadSync: function(req, user, callback) {
    switch (Number(req.status)) {
    case 401:
      delete this._sync[user];
      var msg = "Authentication failed during syncing.";
      EchofonUtils.log(msg);
      EchofonUtils.notifyObservers("internalError", msg);
      break;

    case 205:
      delete this._sync[user].key;
      EchofonUtils.notifyObservers("alertMessage", "Can not sync (User key is not found). Sync is disabled.");
      break;

    case 200:
      var d = new Date();
      var ld = new Date(this._lastSync);
      EchofonUtils.debug("Synced " + user + " on " + d.toTimeString() + " (last sync " + ld.toTimeString() + ": " + req.responseText);
      try {
        var resp = JSON.decode(req.responseText);
      }
      catch (e) {
        EchofonUtils.log('Failed to decode JSON data: ' + e.message);
        break;
      }

      this._sync[user]['config'] = resp['config'];
      this._sync[user]['sync'] = resp['sync_str'];
      if (!this._sync[user].mute || !this._sync[user].mute.updated_at || (resp['mute'].updated_at > this._sync[user].mute.updated_at)) {
        this._sync[user].mute = resp['mute'];
      }
      this._lastSync = Date.now();
      EchofonUtils.notifyComponents("updateSyncData", {user_id:user, data:this._sync[user]});
      break;

    case 503:
      EchofonUtils.log("Echofon sync server is under maintenance.");
      break;

    default:
      var msg = "Unknown error occured during syncing. (" + req.status + ")";
      EchofonUtils.log(msg);
      EchofonUtils.notifyObservers("internalError", msg);
      break;
    }
    this.save();

    if (callback) {
      callback(user);
    }
  },

  onErrorSync: function(req, user, callback) {
    EchofonUtils.log("Failed to sync. Network error.");
    EchofonUtils.notifyObservers("internalError", "Failed to sync. Network error.");

    if (callback) {
      callback();
    }
  },

  save: function() {
    if (this._sync) {
      var val = JSON.encode(this._sync);
      EchofonUtils.pref().setCharPref("sync", val);
    }
  },

  destroy: function() {
    var obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    obs.removeBbserver(this, "quit-application-granted");

    var idleService = Cc["@mozilla.org/widget/idleservice;1"].getService(Ci.nsIIdleService);
    idleService.removeIdleObserver(this, IDLE_TIME_INTERVAL);

    this._syncTimer.cancel();
  },

  observe: function(subject, topic, data) {

    switch (topic) {
    case "idle":
      this._isIdle = 1;
      break;

    case "back":
      this._isIdle = 0;
      if (Date.now() - this._lastSync > this.syncInterval()) {
        this.resetTimer();
        this.getSyncPoint(EchofonUtils.pref().getIntPref("activeUserId"));
      }
      break;

    case "quit-application-granted":
      this.destroy();
      break;

    }
  }
}

var gSync = null;

EchofonSync.instance = function() {
  if (gSync == null) {
    gSync = new EchofonSync();
  }
  return gSync;
}
