var EXPORTED_SYMBOLS = ["PhotoBackend"];

function PhotoBackend(url)
{
  this._url = url;
  this.detectService();
}

const services = ["TwitPic", "Plixi", "yFrog", "Instagram", "Flickr", "YouTube", "TwitVid"];

const PHOTO_LOAD_DELAY       = 300; // ms
const INSTAGRAM_OEM_ENDPOINT = "http://instagr.am/api/v1/oembed/";
const FLICKR_OEM_ENDPOINT    = "http://www.flickr.com/services/oembed/";
const Cc = Components.classes;
const Ci = Components.interfaces;

PhotoBackend.prototype = {
  _url: null,
  _service: null,

  isPhotoURL: function() {
    return this._service;
  },

  detectService: function() {
    var ret = this._url.match('https?://'+
                              '((twitpic\.com/[A-Za-z0-9]+)|'+
                              '((?:plixi\.com/p/\\d+|lockerz\.com/s/\\d+))|'+
                              '(yfrog\.com/\\w+)|'+
                              '(instagr\.am/p/[-_A-Za-z0-9]+/)|'+
                              '(flic\.kr/p/[-_A-Za-z0-9]|(?:www\.)?flickr\.com/photos/[^/]+/\\d+/.*)|'+
                              '((?:www\.)?youtube\.com/\\S+|youtu\.be/\\S+)|'+
                              '((?:www\.)?twitvid\.com/\\w+)'+
                              ')');
    if (ret) {
      for (var i = 2; i <= 8; ++i) {
        if (ret[i]) {
          this._service = services[i-2];
        }
      }
    }
    if (this._service == "YouTube" || this._service == "TwitVid") {
      this.media = "Video";
    }
    else {
      this.media = "Photo";
    }
    return this._service;
  },

  thumbnailURL: function(target_elem) {
    var url = this._url;

    switch (this._service) {
    case "TwitPic":
      return url.replace(/http:\/\/twitpic\.com\/(\w+)/, 'http://twitpic.com/show/mini/$1');
    case "Plixi":
      return "http://api.plixi.com/api/tpapi.svc/json/imagefromurl?size=thumbnail&url=" + url
        + "&TPAPIKEY=50483493-2f21-48e0-b3fb-bc06b631b6dc";
    case "yFrog":
      return url + ".th.jpg";
    case "YouTube":
      var vid = null;
      if (url.match("youtu.be/(.*)")) {
        vid = RegExp.$1;
      }
      else {
        var URI = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService).newURI(url, null, null);
        try {
          var aURL = URI.QueryInterface(Ci.nsIURL);
        }
        catch(e) {
          return "";
        }
        var queries = aURL.query.split('&');
        for (var i in queries) {
          var q = queries[i].split('=');
          if (q[0] == 'v') {
            vid = q[1];
            break;
          }
        }
      }
      if (vid) {
        return "http://img.youtube.com/vi/" + vid + "/default.jpg";
      }

    case "TwitVid":
      return url.replace(/http:\/\/(?:www\.)?twitvid\.com\/(\w+)/, "http://images.twitvid.com/$1.jpg");

    case "Instagram":
//      return url + "/media/?size=t";
      var oc = new OembedClient(INSTAGRAM_OEM_ENDPOINT, target_elem);
      return oc.getImageUrlForPageUrl(url, 200, target_elem);

    case "Flickr":
      if (url.match('flickr.com')) {
        var oc = new OembedClient(FLICKR_OEM_ENDPOINT, target_elem);
        return oc.getImageUrlForPageUrl(url, 75, target_elem);
      }
      else {
        return url.replace(/http:\/\/flic.kr\/p\/(\w+)/, "http://flic.kr/p/img/$1.jpg");
      }
    }
    return "";
  },

  photoURL: function(target_elem) {
    var url = this._url;

    switch (this._service) {
    case "TwitPic":
      return url.replace(/http:\/\/twitpic\.com\/(\w+)/, 'http://twitpic.com/show/large/$1');
    case "Plixi":
      return "http://api.plixi.com/api/tpapi.svc/json/imagefromurl?size=medium&url=" + url
        + "&TPAPIKEY=50483493-2f21-48e0-b3fb-bc06b631b6dc";
    case "yFrog":
      return url + ":iphone";
    case "YouTube":
      return "";
    case "TwitVid":
      return "";

    case "Instagram":
//      return url + "/media/?size=l";
      var oc = new OembedClient(INSTAGRAM_OEM_ENDPOINT, target_elem);
      return oc.getImageUrlForPageUrl(url, 600, target_elem);

    case "Flickr":
      if (url.match('flickr.com')) {
        var oc = new OembedClient(FLICKR_OEM_ENDPOINT, target_elem);
        return oc.getImageUrlForPageUrl(url, 500, target_elem);
      }
      else {
        return url.replace(/http:\/\/flic.kr\/p\/(\w+)/, "http://flic.kr/p/img/$1_m.jpg");
      }
    }
    return "";
  }
};

gOembededResult = {};

function OembedClient(api_end_point, target_elem)
{
  this.endpoint = api_end_point;
  this.target = target_elem;
}

OembedClient.prototype = {
  getImageUrlForPageUrl: function(url, resolution) {

    this.key = url + resolution;
    if (gOembededResult[this.key]) {
      return gOembededResult[this.key];
    }

    var params = {url:url,
                  format:"json",
                  maxwidth:resolution,
                  maxheight:resolution};
    var paramArr = [];
    for (var key in params) {
      paramArr.push(key + "=" + params[key]);
    }
    var requestURL = this.endpoint + "?" + paramArr.join('&');

    var req = Cc["@echofon.com/echofon-http-request;1"].createInstance(Ci.nsIEchofonHttpRequest);
    this.req = req;
    req.setURL(requestURL);
    var target = this;
    req.onload    = function() {target.onLoadOembed(req)};
    req.onerror   = function() {target.onErrorOembed()};
    req.ontimeout = function() {target.onTimeoutOembed()};
    gOembededOperation.add(this);

    return "";
  },

  start: function() {
    this.req.asyncOpen();
  },

  onLoadOembed: function(req) {
    var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);
    try {
      var data = JSON.decode(req.responseText);
      gOembededResult[this.key] = data.url;
      if (this.target.loadImage) this.target.loadImage(data.url);
    }
    catch (e) {
      Components.utils.reportError(e.message);
      Components.utils.reportError(req.responseText);
    }
    this.operation.dequeue();
  },

  onErrorOembed: function() {
    this.operation.dequeue();
  },

  onTimeoutOembed: function() {
    this.operation.dequeue();
  },
};

function OembedOperation()
{
  this.queue = new Array();
}

OembedOperation.prototype = {
  add: function(client) {
    client.operation = this;
    this.queue.push(client);
    if (this.queue.length == 1) {
      client.start();
    }
  },

  dequeue: function() {
    this.queue.shift();
    if (this.queue.length > 0) {
      var obj = this;

      this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
      this.timer.initWithCallback({
        notify: function() {
          obj.start();
        }
      },
      PHOTO_LOAD_DELAY,
      Ci.nsITimer.TYPE_ONE_SHOT);
    }
  },

  start: function() {
    var client = this.queue[0];
    client.start();
  },
};

gOembededOperation = new OembedOperation();
